from wxkit.services.openweather import OpenWeatherService
from wxkit import models


def test_init():
    c = models.Credential(credential={"appid": "*****"})
    s = OpenWeatherService(c)
    assert s



import functools
import os
import re
import requests
import time
from urllib.parse import urlencode
from random import randint
from requests import Session
from requests.exceptions import HTTPError
from requests.packages import urllib3

from wxkit.utils.logger import get_logger


RETRY_TIMES = 3
BACKOFF_FACTOR = 1
STATUS_TO_RETRY = (429, 500, 502, 504)
DEFAULT_TIMEOUT = (30, 120)
COMPOSE_PATH_PATTERN = re.compile("([^/]).*")


class HTTP_METHOD:
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


class RetryException(Exception):
    """Specified exception to retry request."""

    def __init__(self, message):
        super().__init__(message)


def is_ssl_verification():
    return False if os.environ.get("CURL_CA_BUNDLE") == "" else True


def retry(*exceptions):
    """Decorate an async function to execute it a few times before giving up.
    Hopes that problem is resolved by another side shortly.
    Args:
        exceptions (tuple): The exceptions expected during function execution
    """

    logger = get_logger()

    def accumulate_cooldown(retry_times, backoff_factor=None):
        backoff_factor = backoff_factor or BACKOFF_FACTOR
        return backoff_factor * (2 ** retry_times) + (randint(1, 999)) * 0.001

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _count = 0
            _errors = None
            while _count < RETRY_TIMES:
                try:
                    return func(*args, **kwargs)
                except exceptions as err:
                    _errors = err
                    if getattr(_errors, "response", None) is not None:
                        if err.response.status_code not in STATUS_TO_RETRY:
                            break

                    _count += 1
                    time.sleep(
                        accumulate_cooldown(_count, kwargs.get("backoff_factor"))
                    )

            logger.error(f"Retry exhausted error: {str(_errors)}")
            if _errors:
                raise _errors

        return wrapper

    return decorator


@retry(RetryException, HTTPError)
def retry_request(
    method,
    url,
    headers=None,
    files=None,
    data=None,
    json=None,
    params=None,
    auth=None,
    cookies=None,
    hooks=None,
    timeout=DEFAULT_TIMEOUT,
    verify=None,
    session=None,
    proxies=None,
    backoff_factor=BACKOFF_FACTOR,
    jsonify=True,
    object_hook=None,
    **kwargs,
):
    """
    Args:
        method: HTTP request method.
        url: HTTP request url.
        session: The session is used to send HTTP request.
        proxies: The proxy configs are applied to session.
        timeout: The specified timeout for HTTP response.
        jsonify: By default it is True, return response content as dictionary.
            If False, return HTTP response object.
        backoff_factor (float): The factor is used to calculate cooldown for retry.
        object_hook: The hook is used for json.loads if jsonify is True.
        error_handler: The function is used to handle raised exception and accept
            the exception object(maybe None) as first args of function.
    """

    logger = get_logger()
    session = session or Session()
    if isinstance(proxies, dict):
        session.proxies.update(proxies)

    req = requests.Request(
        method=method,
        url=url,
        headers=headers,
        files=files,
        data=data,
        json=json,
        params=params,
        auth=auth,
        cookies=cookies,
        hooks=hooks,
    ).prepare()
    _msg = f"[{method}] {url}?{urlencode(params)}" if params else f"[{method}] {url}"
    logger.info(f"Prepare to request. {_msg}")
    resp_data = None
    verify = verify if isinstance(verify, bool) else is_ssl_verification()
    if verify is False:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    _st = time.time()
    resp = session.send(req, timeout=timeout, verify=verify, **kwargs)
    resp.raise_for_status()
    resp_data = resp.json(object_hook=object_hook) if jsonify else resp
    _et = time.time() - _st
    logger.info(f"Request has finished and costs {_et:f} (s). {_msg}",)

    return resp_data


def compose_url_path(*args):
    # Make sure to trim leading slash if the path args start with.
    _args = [str(a) for a in args]
    _args = (
        _match.group() for _match in map(COMPOSE_PATH_PATTERN.search, _args) if _match
    )
    url_path = os.path.join(*_args)
    return url_path



import logging
import threading
import wxkit
from collections import defaultdict


_LOGGERS = defaultdict(dict)


def get_logger():
    tid = threading.get_ident()
    if _LOGGERS[tid]:
        return _LOGGERS[tid]

    _logger = logging.getLogger(name=wxkit.__name__)
    _logger.setLevel("DEBUG")
    if not _logger.handlers:
        formatter = logging.Formatter(
            "[%(asctime)s][%(levelname)s][%(filename)s][%(funcName)s][%(levelno)s]"
            " - %(message)s"
        )
        _hdr = logging.StreamHandler()
        _hdr.setFormatter(formatter)
        _logger.addHandler(_hdr)

    _LOGGERS[tid] = _logger
    return _logger



from wxkit.services import WeatherService
from wxkit.models import ErrorModel, Wind, Temperature, Station, WeatherPoint
from wxkit.utils.http_request import retry_request, HTTP_METHOD
from wxkit.utils.mathlibs import Vector
from wxkit.utils.convertor import (
    fahrenheit_from_kelvin,
    celsius_from_fahrenheit,
)


class ForecastModel:
    AROME = "arome"
    ICONEU = "iconEu"
    GFS = "gfs"
    WAVEWATCH = "wavewatch"
    NAMCONUS = "namConus"
    NAMHAWAII = "namHawaii"
    NAMALASKA = "namAlaska"
    GEOS5 = "geos5"


class ForecastParameters:
    AROME = [
        "temp",
        "dewpoint",
        "precip",
        "convPrecip",
        "wind",
        "windGust",
        "cape",
        "ptype",
        "lclouds",
        "mclouds",
        "hclouds",
        "rh",
    ]
    ICONEU = [
        "temp",
        "dewpoint",
        "precip",
        "convPrecip",
        "snowPrecip",
        "wind",
        "windGust",
        "cape",
        "ptype",
        "lclouds",
        "mclouds",
        "hclouds",
        "rh",
        "gh",
        "pressure",
    ]
    GFS = [
        "temp",
        "dewpoint",
        "precip",
        "convPrecip",
        "snowPrecip",
        "wind",
        "windGust",
        "cape",
        "ptype",
        "lclouds",
        "mclouds",
        "hclouds",
        "rh",
        "gh",
        "pressure",
    ]
    WAVEWATCH = ["waves", "windWaves", "swell1", "swell2"]
    NAMCONUS = [
        "temp",
        "dewpoint",
        "precip",
        "convPrecip",
        "snowPrecip",
        "wind",
        "windGust",
        "cape",
        "ptype",
        "lclouds",
        "mclouds",
        "hclouds",
        "rh",
        "pressure",
    ]
    NAMHAWAII = [
        "temp",
        "dewpoint",
        "precip",
        "convPrecip",
        "snowPrecip",
        "wind",
        "windGust",
        "cape",
        "ptype",
        "lclouds",
        "mclouds",
        "hclouds",
        "rh",
        "pressure",
    ]
    NAMALASKA = [
        "temp",
        "dewpoint",
        "precip",
        "convPrecip",
        "snowPrecip",
        "wind",
        "windGust",
        "cape",
        "ptype",
        "lclouds",
        "mclouds",
        "hclouds",
        "rh",
        "pressure",
    ]
    GEOS5 = ["so2sm", "dustsm", "cosc"]


ForecastLevels = [
    "surface",
    "1000h",
    "950h",
    "925h",
    "900h",
    "850h",
    "800h",
    "700h",
    "600h",
    "500h",
    "400h",
    "300h",
    "200h",
    "150h",
]


class WindtyService(WeatherService):
    ENDPOINT = "https://api.windy.com/api/point-forecast/v2"

    def __init__(self, credential):
        super().__init__(credential)
        self._payload = dict(key=self.credential.key)

    def _modelize(self, data, location, level):
        wxs = []
        for idx, time in enumerate(data["ts"]):
            temp_k = data[f"temp-{level}"][idx]
            temp_f = fahrenheit_from_kelvin(temp_k)
            temp_c = celsius_from_fahrenheit(temp_f)
            temperature = Temperature(temp_avg=round(temp_c, 2))
            wind_x = data[f"wind_u-{level}"][idx]
            wind_y = data[f"wind_v-{level}"][idx]
            wind_gust = round(data["gust-surface"][idx])
            wind_vector = Vector(wind_x, wind_y)
            wind = Wind(
                speed=wind_gust,
                degree=wind_vector.degree(),
            )
            pressure = round(data["pressure-surface"][idx] / 100, 2)
            humidity = round(data["rh-surface"][idx], 2)
            clouds = (
                data["lclouds-surface"][idx]
                if pressure >= 800.0
                else data["mclouds-surface"][idx]
                if 800.0 > pressure >= 450.0
                else data["hclouds-surface"][idx]
            ) / 100
            clouds = round(clouds, 2)
            snow = round(data["past3hsnowprecip-surface"][idx], 2)
            rain = round(data["past3hconvprecip-surface"][idx], 2)
            weather_point = WeatherPoint(
                station=Station(lat=location.lat, lon=location.lon, name=""),
                status="",
                description="",
                icon="",
                temp=temperature,
                pressure=pressure,
                humidity=humidity,
                rain=rain,
                wind=wind,
                clouds=clouds,
                snow=snow,
                time=time,
                raw_data=data,
            )
            wxs.append(weather_point)
        return wxs

    def get_forecast(self, location):
        model = ForecastModel.GFS
        level = ForecastLevels[0]
        self._payload.update(
            dict(
                lat=location.lat,
                lon=location.lon,
                model=model,
                levels=[level],
                parameters=getattr(ForecastParameters, model.upper()),
            )
        )
        try:
            resp_data = retry_request(
                HTTP_METHOD.POST, self.ENDPOINT, json=self._payload
            )
        except Exception as e:
            return self.handle_error(e)

        wxs = self._modelize(resp_data, location, level)
        return wxs

    def get_current_weather_by_location(self, location):
        wxs = self.get_forecast(location)
        if isinstance(wxs, ErrorModel):
            return wxs

        return wxs[0]

    def get_forecast_weather_by_location(self, location, period=3):
        return self.get_forecast(location)

    def handle_error(self, exception):
        error = ErrorModel()
        if getattr(exception, "response", None) is not None:
            error.code = exception.response.status_code
            error.message = exception.response.text
        else:
            error.message = str(exception)

        return error



import math
from array import array


class Vector:
    __slots__ = ("__x", "__y")
    typecode = "d"

    def __init__(self, x, y):
        self.__x = float(x)
        self.__y = float(y)

    @property
    def x(self):
        return self.__x

    @property
    def y(self):
        return self.__y

    def __iter__(self):
        return (i for i in (self.x, self.y))

    def __repr__(self):
        class_name = type(self).__name__
        return "{}({!r}, {!r})".format(class_name, *self)

    def __str__(self):
        return str(tuple(self))

    def __bytes__(self):
        return bytes([ord(self.typecode)]) + bytes(array(self.typecode, self))

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __abs__(self):
        return math.hypot(self.x, self.y)

    def __bool__(self):
        return bool(abs(self))

    @classmethod
    def frombytes(cls, octets):
        typecode = chr(octets[0])
        memv = memoryview(octets[1:]).cast(typecode)
        return cls(*memv)

    def quardant(self):
        return (
            1
            if self.x > 0 and self.y > 0
            else 2
            if self.x < 0 and self.y > 0
            else 3
            if self.x > 0 and self.y < 0
            else 4
            if self.x < 0 and self.y < 0
            else 0
        )

    def radian(self):
        return math.atan2(self.y, self.x)

    def angle(self):
        return math.degrees(self.radian())

    def degree(self):
        deg = self.angle()
        return deg if deg >= 0.0 else deg + math.degrees(2 * math.pi)

    def magnitude(self):
        return math.sqrt(math.pow(self.x, 2) + math.pow(self.y, 2))

    def __format__(self, fmt_spec=""):
        if fmt_spec.endswith("p"):
            fmt_spec = fmt_spec[:-1]
            coords = (abs(self), self.radian())
            outer_fmt = "<{}, {}>"
        else:
            coords = self
            outer_fmt = "({}, {})"
        components = (format(c, fmt_spec) for c in coords)
        return outer_fmt.format(*components)

    def __hash__(self):
        return hash(self.x) ^ hash(self.y)
